import csv

from glotter.settings import Settings
from glotter.source import get_sources


def report(args):
    filepath = args.output
    reporter = Reporter()
    if filepath is not None:
        reporter.write_csv(filepath)
    else:
        reporter.write_stdout()


class Reporter:
    def __init__(self):
        self._projects = sorted([p.display_name for p in Settings().projects.values()])
        self._language_stats = self._collect_language_stats()
        self._languages = sorted(self._language_stats.keys())

    @staticmethod
    def _get_project_display_name(key):
        return Settings().projects[key].display_name

    def _collect_language_stats(self):
        language_stats = {}
        sources_by_type = get_sources(Settings().source_root)

        for project, sources in sources_by_type.items():
            display_name = self._get_project_display_name(project)
            for source in sources:
                if source.language not in language_stats:
                    language_stats[source.language] = {p: "" for p in self._projects}
                    language_stats[source.language]["Name"] = source.language
                language_stats[source.language][display_name] = f"{source.name}{source.extension}"

        return language_stats

    def write_csv(self, filepath=None):
        filepath = filepath or "glotter-report.csv"
        with open(filepath, "w", newline="", encoding="utf-8") as csv_file:
            languages = sorted(self._language_stats.keys())
            projects = ["Name"] + self._projects
            writer = csv.DictWriter(csv_file, fieldnames=projects)
            writer.writeheader()
            for language in languages:
                writer.writerow(self._language_stats[language])
            print(f"Report written to {filepath}")

    def write_stdout(self):
        column_widths = self._get_column_widths()
        lang_width = max(len(lang) for lang in self._languages)

        header_line = f"| {{0:<{lang_width}}} | ".format("Language")
        header_underline = "| " + ("-" * lang_width) + " | "
        column_template = f"| {{Name:<{lang_width}}} | "
        for project in self._projects:
            header_line += f"{{0:<{column_widths[project]}}} | ".format(project)
            header_underline += ("-" * column_widths[project]) + " | "
            column_template += f"{{{project}:<{column_widths[project]}}} | "

        print(header_line)
        print(header_underline)
        for lang in self._languages:
            print(column_template.format(**self._language_stats[lang]))

    def _get_column_widths(self):
        column_widths = {}
        for project in self._projects:
            max_len = len(project)
            for lang in self._languages:
                if project in self._language_stats[lang]:
                    entry_len = len(self._language_stats[lang][project])
                    max_len = max(max_len, entry_len)

            column_widths[project] = max_len
        return column_widths



import os
from functools import lru_cache

import yaml

from glotter import testinfo
from glotter.containerfactory import ContainerFactory
from glotter.settings import Settings
from glotter.utils import error_and_exit

BAD_SOURCES = "__bad_sources__"


class Source:
    """Metadata about a source file"""

    def __init__(self, name, language, path, test_info_string):
        """Initialize source

        :param name: filename including extension
        :param path: path to the file excluding name
        :param language: the language of the source
        :param test_info_string: a string in yaml format containing testinfo for a directory
        """
        self._name = name
        self._language = language
        self._path = path

        self._test_info = testinfo.TestInfo.from_string(test_info_string, self)

    @property
    def full_path(self):
        """Returns the full path to the source including filename and extension"""
        return os.path.join(self._path, self._name)

    @property
    def path(self):
        """Returns the path to the source excluding name"""
        return self._path

    @property
    def name(self):
        """Returns the name of the source excluding the extension"""
        return os.path.splitext(self._name)[0]

    @property
    def language(self):
        """Returns the language of the source"""
        return self._language

    @property
    def extension(self):
        """Returns the extension of the source"""
        return os.path.splitext(self._name)[1]

    @property
    def test_info(self):
        """Returns parsed TestInfo object"""
        return self._test_info

    def __repr__(self):
        return f"Source(name: {self.name}, path: {self.path})"

    def build(self, params=""):
        if self.test_info.container_info.build is not None:
            command = f"{self.test_info.container_info.build} {params}"
            result = self._container_exec(command)
            if result[0] != 0:
                raise RuntimeError(
                    f'unable to build using cmd "{self.test_info.container_info.build} {params}":\n'
                    f"{result[1].decode('utf-8')}"
                )

    def run(self, params=None):
        """
        Run the source and return the output

        :param params: input passed to the source as it's run
        :return: the output of running the source
        """
        params = params or ""
        command = f"{self.test_info.container_info.cmd} {params}"
        result = self._container_exec(command)
        return result[1].decode("utf-8")

    def exec(self, command):
        """
        Run a command inside the container for a source

        :param command: command to run
        :return:  the output of the command as a string
        """
        result = self._container_exec(command)
        return result[1].decode("utf-8")

    def _container_exec(self, command):
        """
        Run a command inside the container for a source

        :param command: command to run
        :return:  the exit code and output of the command
        """
        container = ContainerFactory().get_container(self)
        return container.exec_run(
            cmd=command,
            detach=False,
            workdir="/src",
        )

    def cleanup(self):
        ContainerFactory().cleanup(self)


@lru_cache
def get_sources(path, check_bad_sources=False):
    """
    Walk through a directory and create Source objects

    :param path: path to the directory through which to walk
    :param check_bad_source: if True, check for bad source filenames. Default is False
    :return: a dict where the key is the ProjectType and the value is a list of all the
        Source objects of that project. If check_bad_source is True,
        the BAD_SOURCES key contains a list of invalid paths relative to the current
        working directory
    """
    sources = {k: [] for k in Settings().projects}
    orig_path = path
    if check_bad_sources:
        sources[BAD_SOURCES] = []

    for root, _, files in os.walk(path):
        path = os.path.abspath(root)
        if "testinfo.yml" in files:
            with open(os.path.join(path, "testinfo.yml"), "r", encoding="utf-8") as file:
                test_info_string = file.read()
            folder_info = testinfo.FolderInfo.from_dict(yaml.safe_load(test_info_string)["folder"])
            folder_project_names = folder_info.get_project_mappings(include_extension=True)
            for project_type, project_name in folder_project_names.items():
                if project_name in files:
                    source = Source(project_name, os.path.basename(path), path, test_info_string)
                    sources[project_type].append(source)

            if check_bad_sources:
                invalid_filenames = set(files) - (
                    set(folder_project_names.values()) | {"testinfo.yml", "README.md"}
                )
                sources[BAD_SOURCES] += [
                    os.path.join(os.path.relpath(path, orig_path), filename)
                    for filename in invalid_filenames
                ]

    return sources


def filter_sources(args, sources):
    """
    Filter sources based language, project, and/or source

    :param args: Arguments indicating what to filter on
    :param sources: a dict where the key is the ProjectType and the value is a list of all the Source objects of that project
    :return: a dict where the key is the ProjectType and the value is a list of all the Source objects of that project
        that match the filter
    """

    if args.project:
        if args.project not in sources:
            error_and_exit(f'No valid sources found for project: "{args.project}"')

        sources = {args.project: sources[args.project]}

    filtered_sources_by_type = {}
    for project_type, sources_by_type in sources.items():
        filtered_sources = [source for source in sources_by_type if _matches_source(args, source)]
        if filtered_sources:
            filtered_sources_by_type[project_type] = filtered_sources

    if not filtered_sources_by_type:
        errors = []
        if args.project:
            errors.append(f'project "{args.project}"')

        if args.language:
            if isinstance(args.language, set):
                errors.append(
                    "languages " + ", ".join(f'"{language}"' for language in sorted(args.language))
                )
            else:
                errors.append(f'language "{args.language}"')

        if args.source:
            errors.append(f'source "{args.source}"')

        if errors:
            error_msg = ", ".join(errors)
            error_and_exit(f"No valid sources found for the following combination: {error_msg}")

    return filtered_sources_by_type


def _matches_source(args, source):
    if args.language:
        if isinstance(args.language, set):
            if source.language.lower() not in args.language:
                return False
        elif source.language.lower() != args.language.lower():
            return False

    return not args.source or f"{source.name}{source.extension}".lower() == args.source.lower()



from typing import List, NoReturn, Optional

from pydantic import ValidationError
from pydantic_core import InitErrorDetails, PydanticCustomError


def get_error_details(msg: str, loc: tuple, input) -> InitErrorDetails:
    return InitErrorDetails(
        type=PydanticCustomError("value_error", msg),
        loc=loc,
        input=input,
    )


def raise_simple_validation_error(cls, msg, input, loc: Optional[tuple] = None) -> NoReturn:
    raise ValidationError.from_exception_data(
        title=cls.__name__,
        line_errors=[get_error_details(msg, loc or (), input)],
    )


def raise_validation_errors(cls, errors: List[InitErrorDetails]) -> NoReturn:
    raise ValidationError.from_exception_data(title=cls.__name__, line_errors=errors)


def validate_str_list(
    cls, values, item_loc: Optional[tuple] = None, raise_exc: bool = True
) -> List[InitErrorDetails]:
    loc = item_loc or ()
    errors = []
    if not isinstance(values, list):
        errors += [get_error_details("Input should be a valid list", loc, values)]
    else:
        errors += [
            get_error_details("Input should be a valid string", loc + (index,), value)
            for index, value in enumerate(values)
            if not isinstance(value, str)
        ]

    if errors and raise_exc:
        raise_validation_errors(cls, errors)

    return errors


def validate_str_dict(
    cls, values, item_loc: Optional[tuple] = None, raise_exc: bool = True
) -> List[InitErrorDetails]:
    loc = item_loc or ()
    errors = []
    if not isinstance(values, dict):
        errors += [get_error_details("Input should be a valid dictionary", loc, values)]
    else:
        errors += sum(
            (
                _get_str_dict_error_details(loc + (key,), key, value)
                for key, value in values.items()
                if not isinstance(key, str) or not isinstance(value, str)
            ),
            [],
        )

    if errors and raise_exc:
        raise_validation_errors(cls, errors)

    return errors


def _get_str_dict_error_details(loc, key, value) -> List[InitErrorDetails]:
    errors = []
    if not isinstance(key, str):
        errors += [get_error_details("Key should be a valid string", loc, key)]

    if not isinstance(value, str):
        errors += [get_error_details("Value should be a valid string", loc, value)]

    return errors



from concurrent.futures import ThreadPoolExecutor

from glotter.containerfactory import ContainerFactory
from glotter.settings import Settings
from glotter.source import filter_sources, get_sources


def download(args):
    def get_key(source):
        return f"{source.test_info.container_info.image}:{source.test_info.container_info.tag}"

    sources_by_type = filter_sources(args, get_sources(Settings().source_root))
    containers = {
        get_key(source): source.test_info.container_info
        for sources in sources_by_type.values()
        for source in sources
    }
    if args.parallel:
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.map(
                lambda source: _download_image_from_source(source, True),
                containers.values(),
            )
    else:
        for container_info in containers.values():
            _download_image_from_source(container_info)

    return containers


def _download_image_from_source(container_info, parallel=False):
    ContainerFactory().get_image(container_info, parallel=parallel)


def remove_images(containers, parallel):
    if parallel:
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.map(ContainerFactory().remove_image, containers.values())
    else:
        for container_info in containers.values():
            ContainerFactory().remove_image(container_info)



from functools import partial
from typing import Annotated, Any, Callable, ClassVar, Dict, List, Optional, Tuple

from pydantic import (
    BaseModel,
    Field,
    ValidationError,
    ValidationInfo,
    field_validator,
    model_validator,
)

from glotter.errors import (
    get_error_details,
    raise_simple_validation_error,
    raise_validation_errors,
    validate_str_dict,
    validate_str_list,
)
from glotter.utils import indent, quote

TransformationScalarFuncT = Callable[[str, str], Tuple[str, str]]
TransformationDictFuncT = Callable[[List[str], str, str], Tuple[str, str]]


class AutoGenParam(BaseModel):
    """Object used to auto-generated a test parameter"""

    name: str = ""
    input: Optional[str] = None
    expected: Any

    @field_validator("expected")
    def validate_expected(cls, value):
        """
        Validate expected value

        :param value: Expected value
        :return: Original expected value
        :raises: :exc:`ValidationError` if invalid expected value
        """

        if isinstance(value, dict):
            if not value:
                raise_simple_validation_error(cls, "Too few items", value)

            if len(value) > 1:
                raise_simple_validation_error(cls, "Too many items", value)

            key, item = tuple(*value.items())
            if key in ("exec", "string"):
                if not isinstance(item, str):
                    raise_simple_validation_error(
                        cls, "Input should be a valid string", item, (key,)
                    )
                if not item:
                    raise_simple_validation_error(cls, "Value must not be empty", item, (key,))
            elif key != "self":
                raise_simple_validation_error(cls, 'Invalid "expected" type', item)
        elif isinstance(value, list):
            validate_str_list(cls, value)
        elif not isinstance(value, str):
            raise_simple_validation_error(
                cls, "Input should be a valid string, list, or dictionary", value
            )

        return value

    def get_pytest_param(self) -> str:
        """
        Get pytest parameter string

        :return: pytest parameter string if name is not empty, empty string otherwise
        """

        if not self.name:
            return ""

        input_param = self.input
        if isinstance(input_param, str):
            input_param = quote(input_param)

        expected_output = self.expected
        if isinstance(expected_output, str):
            expected_output = quote(expected_output)
        elif isinstance(expected_output, dict) and "string" in expected_output:
            expected_output = self.get_constant_variable_name()

        return f"pytest.param({input_param}, {expected_output}, id={quote(self.name)}),\n"

    def get_constant_variable_name(self) -> str:
        """
        Get constant variable name

        :return: constant variable name
        """

        variable_name = ""
        if isinstance(self.expected, dict) and "string" in self.expected:
            variable_name = self.expected["string"].upper()

        return variable_name


def _append_method_to_actual(method: str, actual_var: str, expected_var) -> Tuple[str, str]:
    return f"{actual_var}.{method}()", expected_var


def _append_method_to_expected(method: str, actual_var: str, expected_var: str) -> Tuple[str, str]:
    return actual_var, f"{expected_var}.{method}()"


def _remove_chars(values: List[str], actual_var: str, expected_var: str) -> Tuple[str, str]:
    for value in values:
        actual_var += f'.replace({quote(value)}, "")'

    return actual_var, expected_var


def _strip_chars(values: List[str], actual_var: str, expected_var: str) -> Tuple[str, str]:
    for value in values:
        actual_var += f".strip({quote(value)})"

    return actual_var, expected_var


def _unique_sort(actual_var, expected_var):
    return f"sorted(set({actual_var}))", f"sorted(set({expected_var}))"


class AutoGenTest(BaseModel):
    """Object used to auto-generated a test"""

    name: Annotated[str, Field(strict=True, min_length=1, pattern="^[a-zA-Z][0-9a-zA-Z_]*$")]
    requires_parameters: bool = False
    inputs: Annotated[List[str], Field(strict=True, min_length=1)] = Field(
        ["Input"], validate_default=True
    )
    params: Annotated[List[AutoGenParam], Field(strict=True, min_length=1)] = Field(
        None, validate_default=True
    )
    strings: Dict[str, str] = {}
    transformations: List[Any] = []

    SCALAR_TRANSFORMATION_FUNCS: ClassVar[Dict[str, TransformationScalarFuncT]] = {
        "strip": partial(_append_method_to_actual, "strip"),
        "splitlines": partial(_append_method_to_actual, "splitlines"),
        "lower": partial(_append_method_to_actual, "lower"),
        "any_order": _unique_sort,
        "strip_expected": partial(_append_method_to_expected, "strip"),
        "splitlines_expected": partial(_append_method_to_expected, "splitlines"),
    }
    DICT_TRANSFORMATION_FUNCS: ClassVar[Dict[str, TransformationDictFuncT]] = {
        "remove": _remove_chars,
        "strip": _strip_chars,
    }

    @field_validator("inputs", mode="before")
    @classmethod
    def validate_inputs(cls, values):
        """
        Validate each input

        :param values: Inputs to validate
        :return: Original inputs
        :raises: :exc:`ValidationError` if input invalid
        """

        validate_str_list(cls, values)
        return values

    @field_validator("params", mode="before")
    @classmethod
    def validate_params(cls, values, info: ValidationInfo):
        """
        Validate each parameter

        :param values: Parameters to validate
        :param info: Test item
        :return: Original parameters
        :raises: :exc:`ValidationError` if project requires parameters but no input, no name,
            or empty name. Also, raised if no expected output
        """

        errors = []
        field_is_required = "Field is required when parameters required"

        for index, value in enumerate(values):
            if info.data.get("requires_parameters"):
                if not isinstance(value, dict):
                    errors.append(
                        get_error_details("Input should be a valid dictionary", (index,), value)
                    )
                    continue

                if "name" not in value:
                    errors.append(get_error_details(field_is_required, (index, "name"), value))
                elif isinstance(value["name"], str) and not value["name"]:
                    errors.append(
                        get_error_details(
                            "Value must not be empty when parameters required",
                            (index, "name"),
                            value,
                        )
                    )

                if "input" not in value:
                    errors.append(get_error_details(field_is_required, (index, "input"), value))

            if "expected" not in value:
                errors.append(get_error_details(field_is_required, (index, "expected"), value))

        if errors:
            # Collect inner errors
            for index, value in enumerate(values):
                if not isinstance(value, dict):
                    continue

                try:
                    AutoGenParam.model_validate(value)
                except ValidationError as exc:
                    for err in exc.errors():
                        loc = (index,) + tuple(err.get("loc", ()))
                        msg = err.get("msg") or str(err.get("type", "value_error"))
                        input_val = err.get("input", value)
                        errors.append(get_error_details(msg, loc, input_val))

            raise_validation_errors(cls, errors)

        return values

    @field_validator("strings", mode="before")
    @classmethod
    def validate_strings(cls, values):
        """
        Validate each string

        :param values: Strings to validate
        :return: Original strings
        :raises: :exc:`ValidationError` if strings invalid
        """

        validate_str_dict(cls, values)
        return values

    @field_validator("transformations", mode="before")
    @classmethod
    def validate_transformation(cls, values):
        """
        Validate each transformation

        :param values: Transformations to validate
        :return: Original values
        :raises: :exc:`ValidationError` if Invalid transformation
        """

        if not isinstance(values, list):
            raise_simple_validation_error(cls, "Input should be a valid list", values)

        errors = []
        for index, value in enumerate(values):
            if isinstance(value, str):
                if value not in cls.SCALAR_TRANSFORMATION_FUNCS:
                    errors.append(
                        get_error_details(f'Invalid transformation "{value}"', (index,), value)
                    )
            elif isinstance(value, dict):
                key = str(*value)
                if key not in cls.DICT_TRANSFORMATION_FUNCS:
                    errors.append(
                        get_error_details(f'Invalid transformation "{key}"', (index,), value)
                    )
                else:
                    errors += validate_str_list(cls, value[key], (index, key), raise_exc=False)
            else:
                errors.append(
                    get_error_details(
                        "Input should be a valid string or dictionary", (index,), value
                    )
                )

        if errors:
            raise_validation_errors(cls, errors)

        return values

    @model_validator(mode="after")
    def validate_test_strings(self):
        """
        Validate each test string
        """

        errors = []
        if self.requires_parameters:
            for index, param in enumerate(self.params):
                loc = ("params", index, "expected", "string")
                expected = param.expected
                if (
                    isinstance(param.expected, dict)
                    and "string" in expected
                    and expected["string"] not in self.strings
                ):
                    expected_string = expected["string"]
                    errors.append(
                        get_error_details(
                            f"Refers to a non-existent string {expected_string}",
                            loc=loc,
                            input=expected_string,
                        )
                    )

        if errors:
            raise_validation_errors(self.__class__, errors)

        return self

    def transform_vars(self) -> Tuple[str, str]:
        """
        Transform variables using the specified transformations

        :return: Transformed actual and expected variables
        """

        actual_var = "actual"
        expected_var = "expected"
        for transfomation in self.transformations:
            if isinstance(transfomation, str):
                actual_var, expected_var = self.SCALAR_TRANSFORMATION_FUNCS[transfomation](
                    actual_var, expected_var
                )
            else:
                key, item = tuple(*transfomation.items())
                actual_var, expected_var = self.DICT_TRANSFORMATION_FUNCS[key](
                    item, actual_var, expected_var
                )

        return actual_var, expected_var

    def get_constant_variables(self) -> str:
        """
        Get contant variables

        :return: constant variables
        """

        constant_variables = {}
        if self.requires_parameters:
            for param in self.params:
                variable_name = param.get_constant_variable_name()
                if variable_name and variable_name not in constant_variables:
                    constant_variables[variable_name] = self.strings[param.expected["string"]]

        return "".join(f"{name} = {quote(value)}\n" for name, value in constant_variables.items())

    def get_pytest_params(self) -> str:
        """
        Get pytest parameters

        :return: pytest parameters
        """

        if not self.requires_parameters:
            return ""

        pytest_params = "".join(
            indent(param.get_pytest_param(), 8) for param in self.params
        ).strip()
        return f"""\
@pytest.mark.parametrize(
    ("in_params", "expected"),
    [
        {pytest_params}
    ]
)
"""

    def get_test_function_and_run(self, project_name_underscores: str) -> str:
        """
        Get test function and run command

        :param project_name_underscores: Project name with underscores between each word
        :return: Test function and run command
        """

        func_params = ""
        run_param = ""
        if self.requires_parameters:
            func_params = "in_params, expected, "
            run_param = "params=in_params"

        return f"""\
def test_{self.name}({func_params}{project_name_underscores}):
    actual = {project_name_underscores}.run({run_param})
"""

    def get_expected_output(self, project_name_underscores: str) -> str:
        """
        Get test code that gets the expected output

        :param project_name_underscores: Project name with underscores between each word
        :return: Test code that gets the expected output

        """

        if self.requires_parameters:
            return ""

        expected_output = self.params[0].expected
        if isinstance(expected_output, str):
            expected_output = quote(expected_output)
        elif isinstance(expected_output, dict):
            return _get_expected_file(project_name_underscores, expected_output)

        return f"expected = {expected_output}\n"

    def generate_test(self, project_name_underscores: str) -> str:
        """
        Generate test code

        :param project_name_underscores: Project name with underscores between each word
        :return: Test code
        """

        test_code = "@project_test(PROJECT_NAME)\n"
        test_code += self.get_pytest_params()
        test_code += self.get_test_function_and_run(project_name_underscores)
        test_code += indent(self.get_expected_output(project_name_underscores), 4)
        actual_var, expected_var = self.transform_vars()
        test_code += indent(_get_assert(actual_var, expected_var, self.params[0].expected), 4)
        return test_code


def _get_expected_file(project_name_underscores: str, expected_output: Dict[str, str]) -> str:
    if "exec" in expected_output:
        script = quote(expected_output["exec"])
        return f"expected = {project_name_underscores}.exec({script})\n"

    test_code = f"""\
with open({project_name_underscores}.full_path, "r", encoding="utf-8") as file:
    expected = file.read()
"""

    if "self" in expected_output:
        test_code += """\
diff_len = len(actual) - len(expected)
if diff_len > 0:
    expected += "\\n"
elif diff_len < 0:
    actual += "\\n"
"""

    return test_code


def _get_assert(actual_var: str, expected_var: str, expected_output) -> str:
    if isinstance(expected_output, list):
        return f"""\
actual_list = {actual_var}
expected_list = {expected_var}
assert len(actual_list) == len(expected_list), "Length not equal"
for index in range(len(expected_list)):
    assert actual_list[index] == expected_list[index], f"Item {{index + 1}} is not equal"
"""

    test_code = ""
    if actual_var != "actual":
        test_code += f"actual = {actual_var}\n"

    if expected_var != "expected":
        test_code += f"expected = {expected_var}\n"

    return f"{test_code}assert actual == expected\n"


class AutoGenUseTests(BaseModel):
    """Object used to specify what tests to use"""

    name: str
    search: Annotated[str, Field(strict=True, pattern="^[0-9a-zA-Z_]*$")] = ""
    replace: Annotated[str, Field(strict=True, pattern="^[0-9a-zA-Z_]*$")] = ""

    @model_validator(mode="before")
    @classmethod
    def validate_search_with_replace(cls, values):
        """
        Validate that if either search or replace is specified, both must be specified

        :param values: Values to validate
        :return: Original values
        :raise: `exc`:ValidationError if either search or replace is specified, both are specified
        """

        if "search" in values and "replace" not in values:
            raise_simple_validation_error(
                cls, '"search" item specified without "replace" item', values
            )

        if "search" not in values and "replace" in values:
            raise_simple_validation_error(
                cls, '"replace" item specified without "search" item', values
            )

        return values



import os
from typing import Dict, Optional
from warnings import warn

import yaml
from pydantic import (
    BaseModel,
    Field,
    ValidationError,
    ValidationInfo,
    field_validator,
    model_validator,
)

from glotter.errors import get_error_details, raise_simple_validation_error, raise_validation_errors
from glotter.project import AcronymScheme, Project
from glotter.singleton import Singleton
from glotter.utils import error_and_exit, indent


class Settings(metaclass=Singleton):
    def __init__(self):
        self._project_root = os.getcwd()
        try:
            self._parser = SettingsParser(self._project_root)
        except ValidationError as e:
            error_and_exit(_format_validate_error(e))

        self._projects = self._parser.projects
        self._source_root = self._parser.source_root or self._project_root
        self._test_mappings = {}

    @property
    def projects(self):
        return self._projects

    @property
    def project_root(self):
        return self._project_root

    @property
    def source_root(self):
        return self._source_root

    @source_root.setter
    def source_root(self, value):
        self._source_root = value or self._project_root

    @property
    def test_mappings(self):
        return self._test_mappings

    def get_test_mapping_name(self, project_type):
        mappings = self._test_mappings.get(project_type)
        if mappings:
            return [func.__name__ for func in mappings]
        return []

    def add_test_mapping(self, project_type, func):
        if project_type not in self._projects:
            raise KeyError(f"Project type {project_type} was not found in glotter.yml")

        if project_type not in self._test_mappings:
            self._test_mappings[project_type] = []
        self._test_mappings[project_type].append(func)

    def verify_project_type(self, name):
        return name.lower() in self.projects


def _format_validate_error(validation_error: ValidationError) -> str:
    error_msgs = []
    for error in validation_error.errors():
        error_msgs.append(
            "- "
            + ".".join(
                _format_location_item(location)
                for location in error["loc"]
                if location != "__root__"
            )
            + ":"
        )
        error_msgs.append(indent(error["msg"], 4))

    return "Errors found in the following items:\n" + "\n".join(error_msgs)


def _format_location_item(location) -> str:
    if isinstance(location, int):
        return f"item {location + 1}"

    return str(location)


class SettingsConfigSettings(BaseModel):
    acronym_scheme: AcronymScheme = Field(AcronymScheme.two_letter_limit, validate_default=True)
    yml_path: str
    source_root: Optional[str] = None

    @field_validator("acronym_scheme", mode="before")
    @classmethod
    def get_acronym_scheme(cls, value):
        if isinstance(value, str):
            return value.lower()

        return value

    @field_validator("source_root", mode="after")
    @classmethod
    def get_source_root(cls, value, info: ValidationInfo):
        if os.path.isabs(value):
            return value

        yml_dir = os.path.dirname(info.data["yml_path"])
        return os.path.abspath(os.path.join(yml_dir, value))


class SettingsConfig(BaseModel):
    yml_path: str
    settings: Optional[SettingsConfigSettings] = Field(None, validate_default=True)
    projects: Dict[str, Project] = {}

    @field_validator("settings", mode="before")
    @classmethod
    def get_settings(cls, value, info: ValidationInfo):
        if value is None:
            return {"yml_path": info.data["yml_path"]}

        if isinstance(value, dict):
            return {**value, "yml_path": info.data["yml_path"]}

        return value

    @field_validator("projects", mode="before")
    @classmethod
    def get_projects(cls, value, info: ValidationInfo):
        if not isinstance(value, dict):
            raise_simple_validation_error(cls, "Input should be a valid dictionary", value)

        acronym_scheme = info.data["settings"].acronym_scheme
        for project_name, item in value.items():
            if not isinstance(item, dict):
                break

            value[project_name] = {**item, "acronym_scheme": acronym_scheme}

        return value

    @model_validator(mode="after")
    def validate_projects(self):
        projects = self.projects
        if not isinstance(projects, dict):
            return self

        projects_with_use_tests = {
            project_name: project for project_name, project in projects.items() if project.use_tests
        }

        errors = []
        for project_name, project in projects_with_use_tests.items():
            use_tests_name = project.use_tests.name
            loc = ("projects", project_name, "use_tests")

            # Make sure "use_tests" item refers to an actual project
            if use_tests_name not in projects:
                errors.append(
                    get_error_details(
                        f"Refers to a non-existent project {use_tests_name}",
                        loc=loc,
                        input=use_tests_name,
                    )
                )
            # Make sure one "use_tests" item does not refer to another "use_tests" item
            elif use_tests_name in projects_with_use_tests:
                errors.append(
                    get_error_details(
                        f'Refers to another "use_tests" project {use_tests_name}',
                        loc=loc,
                        input=use_tests_name,
                    )
                )
            # Make sure "use_tests" item refers to a project with tests
            elif not projects[use_tests_name].tests:
                errors.append(
                    get_error_details(
                        f'Refers to project {use_tests_name}, which has no "tests" item',
                        loc=loc,
                        input=use_tests_name,
                    )
                )
            # Otherwise, set the tests that the "use_tests" item refers to with the tests renamed
            else:
                project.set_tests(projects[use_tests_name])

        if errors:
            raise_validation_errors(self.__class__, errors)

        return self


class SettingsParser:
    def __init__(self, project_root):
        self._project_root = project_root
        self._yml_path = None
        self._acronym_scheme = None
        self._projects = None
        self._source_root = None
        self._yml_path = self._locate_yml()

        yml = None
        if self._yml_path is not None:
            yml = self._parse_yml()
        else:
            self._yml_path = project_root
            warn(f'.glotter.yml not found in directory "{project_root}"')

        if yml is None:
            yml = {}

        if not isinstance(yml, dict):
            error_and_exit(".glotter.yml does not contain a dict")

        config = SettingsConfig(**yml, yml_path=self._yml_path)
        self._acronym_scheme = config.settings.acronym_scheme
        self._source_root = config.settings.source_root
        self._projects = config.projects

    @property
    def project_root(self):
        return self._project_root

    @property
    def yml_path(self):
        return self._yml_path

    @property
    def source_root(self):
        return self._source_root

    @property
    def acronym_scheme(self):
        return self._acronym_scheme

    @property
    def projects(self):
        return self._projects

    def _parse_yml(self):
        with open(self._yml_path, "r", encoding="utf-8") as f:
            contents = f.read()

        return yaml.safe_load(contents)

    def _locate_yml(self):
        for root, _, files in os.walk(self._project_root):
            if ".glotter.yml" in files:
                path = os.path.abspath(root)
                return os.path.join(path, ".glotter.yml")

        return None



from enum import Enum, auto
from typing import Annotated, ClassVar, Dict, List, Optional

from pydantic import BaseModel, Field, ValidationInfo, field_validator

from glotter.auto_gen_test import AutoGenTest, AutoGenUseTests
from glotter.errors import raise_simple_validation_error, validate_str_dict, validate_str_list


class NamingScheme(Enum):
    hyphen = auto()
    underscore = auto()
    camel = auto()
    pascal = auto()
    lower = auto()


class AcronymScheme(Enum):
    lower = "lower"
    upper = "upper"
    two_letter_limit = "two_letter_limit"


class Project(BaseModel):
    VALID_REGEX: ClassVar[str] = "^[0-9a-zA-Z]+$"

    words: Annotated[
        List[Annotated[str, Field(min_length=1, pattern=VALID_REGEX, strict=True)]],
        Field(min_length=1, strict=True),
    ]
    requires_parameters: bool = False
    strings: Dict[str, str] = {}
    acronyms: List[Annotated[str, Field(min_length=1, pattern=VALID_REGEX, strict=True)]] = []
    acronym_scheme: AcronymScheme = AcronymScheme.two_letter_limit
    use_tests: Optional[AutoGenUseTests] = None
    tests: Dict[str, AutoGenTest] = {}

    @field_validator("acronyms", mode="before")
    @classmethod
    def get_acronym(cls, values):
        validate_str_list(cls, values)
        return [value.upper() for value in values]

    @field_validator("strings", mode="before")
    @classmethod
    def validate_strings(cls, values):
        validate_str_dict(cls, values)
        return values

    @field_validator("tests", mode="before")
    @classmethod
    def get_tests(cls, value, info: ValidationInfo):
        if not isinstance(value, dict) or not all(
            isinstance(test, dict) for test in value.values()
        ):
            return value

        if info.data.get("use_tests"):
            raise_simple_validation_error(
                cls, '"tests" and "use_tests" items are mutually exclusive', {"use_tests": "..."}
            )

        return {
            test_name: {
                **test,
                "requires_parameters": info.data.get("requires_parameters") or False,
                "strings": info.data.get("strings") or {},
                "name": test_name,
            }
            for test_name, test in value.items()
        }

    def set_tests(self, project: "Project"):
        """
        If there is a "use_tests" item, then set the specified tests, renaming them
        according to the "use_tests" item. The "use_tests" item is then removed

        :params tests: Project with tests to use
        """

        if self.use_tests:
            self.tests = {}
            for test_name_, test in project.tests.items():
                test_name = test_name_.replace(self.use_tests.search, self.use_tests.replace)
                self.tests[test_name] = AutoGenTest(
                    **test.model_dump(exclude={"name"}), name=test_name
                )

            self.requires_parameters = project.requires_parameters
            self.use_tests = None

    @property
    def display_name(self):
        return self._as_display()

    def get_project_name_by_scheme(self, naming):
        """
        gets a project name for a specific naming scheme

        :param naming: the naming scheme
        :return: the project type formatted by the directory's naming scheme
        """
        try:
            return {
                NamingScheme.hyphen: self._as_hyphen(),
                NamingScheme.underscore: self._as_underscore(),
                NamingScheme.camel: self._as_camel(),
                NamingScheme.pascal: self._as_pascal(),
                NamingScheme.lower: self._as_lower(),
            }[naming]
        except KeyError as e:
            raise KeyError(f'Unknown naming scheme "{naming}"') from e

    def _as_hyphen(self):
        return "-".join(self._try_as_acronym(word, NamingScheme.hyphen) for word in self.words)

    def _as_underscore(self):
        return "_".join(self._try_as_acronym(word, NamingScheme.underscore) for word in self.words)

    def _as_camel(self):
        return self.words[0].lower() + "".join(
            self._try_as_acronym(word.title(), NamingScheme.camel) for word in self.words[1:]
        )

    def _as_pascal(self):
        return "".join(
            self._try_as_acronym(word.title(), NamingScheme.pascal) for word in self.words
        )

    def _as_lower(self):
        return "".join(word.lower() for word in self.words)

    def _as_display(self):
        return " ".join(
            self._try_as_acronym(word.title(), NamingScheme.underscore) for word in self.words
        )

    def _is_acronym(self, word):
        return word.upper() in self.acronyms

    def _try_as_acronym(self, word, naming_scheme):
        if self._is_acronym(word):
            if self.acronym_scheme == AcronymScheme.upper:
                return word.upper()
            elif self.acronym_scheme == AcronymScheme.lower:
                return word.lower()
            elif len(word) <= 2 and naming_scheme in [
                NamingScheme.camel,
                NamingScheme.pascal,
            ]:
                return word.upper()

        return word



import yaml
from jinja2 import BaseLoader, Environment

from glotter.project import NamingScheme
from glotter.settings import Settings


class ContainerInfo:
    """Configuration for a container to run for a directory"""

    def __init__(self, image, tag, cmd, build=None):
        """
        Initialize a ContainerInfo

        :param image: the image to run
        :param tag: the tag of the image to run
        :param cmd: the command to run the source inside the container
        :param build: an optional command to run to build the source before running the command
        """
        self._image = image
        self._cmd = cmd
        self._tag = tag
        self._build = build

    @property
    def image(self):
        """Returns the image to run"""
        return self._image

    @property
    def cmd(self):
        """Returns the command to run the source inside the container"""
        return self._cmd

    @property
    def tag(self):
        """Returns the tag of the image to run"""
        return self._tag

    @property
    def build(self):
        """Returns the command to build the source before running it inside the container"""
        return self._build

    @classmethod
    def from_dict(cls, dictionary):
        """
        Create a ContainerInfo from a dictionary

        :param dictionary: the dictionary representing ContainerInfo
        :return: a new ContainerInfo
        """
        image = dictionary["image"]
        tag = dictionary["tag"]
        cmd = dictionary["cmd"]
        build = dictionary["build"] if "build" in dictionary else None
        return ContainerInfo(image=image, tag=tag, cmd=cmd, build=build)

    def __eq__(self, other):
        return (
            self.image == other.image
            and self.cmd == other.cmd
            and self.tag == other.tag
            and self.build == other.build
        )


class FolderInfo:
    """Metadata about sources in a directory"""

    def __init__(self, extension, naming):
        """
        Initialize a FolderInfo

        :param extension: the file extension that is considered as source
        :param naming: the naming scheme for files in the directory
        """
        self._extension = extension
        try:
            self._naming = NamingScheme[naming]
        except KeyError as e:
            raise KeyError(f'Unknown naming scheme: "{naming}"') from e

    @property
    def extension(self):
        """Returns the extension for sources in the directory"""
        return self._extension

    @property
    def naming(self):
        """Returns the naming scheme for the directory"""
        return self._naming

    def get_project_mappings(self, include_extension=False):
        """
        Uses the naming scheme to generate the expected source names in the directory
        and create a mapping from ProjectType to source name

        :param include_extension: whether to include the extension in the source name
        :return: a dict where the key is a ProjectType and the value is the source name
        """
        extension = self.extension if include_extension else ""
        return {
            project_type: f"{project.get_project_name_by_scheme(self.naming)}{extension}"
            for project_type, project in Settings().projects.items()
        }

    def __eq__(self, other):
        return self.extension == other.extension and self.naming == other.naming

    @classmethod
    def from_dict(cls, dictionary):
        """
        Create a FileInfo from a dictionary

        :param dictionary: the dictionary representing FileInfo
        :return: a new FileInfo
        """
        return FolderInfo(dictionary["extension"], dictionary["naming"])


class TestInfo:
    """an object representation of a testinfo file"""

    __test__ = False  # Indicate this is not a test

    def __init__(self, container_info, file_info):
        """
        Initialize a TestInfo object

        :param container_info: ContainerInfo object
        :param file_info: FileInfo object
        """
        self._container_info = container_info
        self._file_info = file_info

    @property
    def container_info(self):
        """Return container info section"""
        return self._container_info

    @property
    def file_info(self):
        """Return file info section"""
        return self._file_info

    @classmethod
    def from_dict(cls, dictionary):
        """
        Create a TestInfo from a dictionary

        :param dictionary: the dictionary representing TestInfo
        :return: a new TestInfo
        """
        return TestInfo(
            container_info=ContainerInfo.from_dict(dictionary["container"]),
            file_info=FolderInfo.from_dict(dictionary["folder"]),
        )

    @classmethod
    def from_string(cls, string, source):
        """
        Create a TestInfo from a string. Modify the string using Jinja2 templating. Then parse it as yaml

        :param string: contents of a testinfo file
        :param source: a source object to use for jinja2 template parsing
        :return: a new TestInfo
        """
        template = Environment(loader=BaseLoader).from_string(string)
        template_string = template.render(source=source)
        info_yaml = yaml.safe_load(template_string)
        return cls.from_dict(info_yaml)

    def __eq__(self, other):
        return self.container_info == other.container_info and self.file_info == other.file_info



import sys

from glotter.settings import Settings
from glotter.source import BAD_SOURCES, get_sources


def check(_args):
    # Get all sources
    all_sources = get_sources(Settings().source_root, check_bad_sources=True)

    # If no bad sources, exit with zero status
    bad_sources = all_sources[BAD_SOURCES]
    if not bad_sources:
        print("All filenames correspond to valid projects")
        sys.exit(0)

    # Show bad sources and exit with non-zero status
    print("The following filenames do not correspond to a valid project:")
    for source in sorted(bad_sources):
        print(f"- {source}")

    sys.exit(1)



from glotter.__main__ import main
from glotter.decorators import project_fixture, project_test
from glotter.settings import Settings
from glotter.test_doc_generator import generate_test_docs

__all__ = ["Settings", "generate_test_docs", "main", "project_fixture", "project_test"]
